#include "binio.h"
#include "binio-internal.h"
#include <stdio.h>
#include <stdint.h>

static
int
get_uint(FILE * const restrict f, uintmax_t * const restrict n, unsigned int const s)
{
	unsigned int i;
	unsigned int const rs = s < sizeof *n ? s : sizeof *n;

	for (i = rs; i < s; i++) {
		int c;
	
		if ((c = getc(f)) == EOF)
			return 1;
		/*
		 * we may have to clip it.  e.g. if uintmax_t is only 64 bits in size but we're trying to read a 128-bit
		 * integer (or whatever), then obviously we can't deal with it.  thus, we throw out the leading bytes that
		 * we can't handle.  e.g. if we have:
		 *	00000000000000003fa4
		 * but we can only handle 8 digits, then we want to throw out the leading zeroes:
		 *      000000000000         => bit bucket
		 * and read in the remainder:
		 *                  00003fa4 => *n
		 *
		 */
	}
	for (*n = 0, i = 0; i < rs; i++) {
		int c;
		
		c = getc(f);
		if (c == EOF)
			return 1;
	
		*n |= (uintmax_t)c << (BYTE_BIT * (s - i - 1));
		/*
		 * start off with *n as all-bits 0, then go through each byte (which may be something different from a char)
		 * and change *n accordingly
		 *
		 */
	}
	
	return 0;
/*
 * get an unsigned uint of size 's' bits from file stream 'f' and store it in *'n'
 *
 */
}

#include <limits.h>

static
int
get_int(FILE * const restrict f, intmax_t * const restrict n, unsigned int const s)
{
	uintmax_t rn;
	unsigned int const padding = sizeof rn * CHAR_BIT - s * BYTE_BIT;
	
	if (get_uint(f, &rn, s))
		return 1;

	*n = rn & ((uintmax_t)1 << ((BYTE_BIT * s) - 1)) ? -(((~rn + 1) << padding) >> padding) : rn;
	
	return 0;
/*
 * get an unsigned int and then bit-play if necessary in order to get it from two's complement into negative native
 *
 */
}

#include <math.h>

static
int
get_float(FILE * const restrict f, long_double * const restrict n, unsigned int const ms, unsigned int const es)
{
	intmax_t mantissa, exponent;
	unsigned int const ms_shift = BYTE_BIT * ms;
	unsigned int const shift = sizeof mantissa * CHAR_BIT < ms_shift ? sizeof mantissa * CHAR_BIT : ms_shift;
	
	if (get_int(f, &mantissa, ms) || get_int(f, &exponent, es))
		return 1;

	*n = ldexpl(mantissa, exponent - (shift - 1));
	
	return 0;
}

#include <stdarg.h>

int
fgetf(FILE * const restrict f, char const * const restrict fmt, ...)
{
	va_list a;
	
	va_start(a, fmt);
	{
	int const r = vfgetf(f, fmt, a);
	va_end(a);
	
	return r;
	}
}

#ifdef BINIO_HAVE_WCHAR

int
fwgetf(FILE * const restrict f, wchar_t const * const restrict fmt, ...)
{
	va_list a;
	
	va_start(a, fmt);
	{
	int const r = vfwgetf(f, fmt, a);
	va_end(a);
	
	return r;
	}
}

#endif

#include <stddef.h>

void
BINIO_assign_int(intmax_t const v, unsigned int const s, va_list * const restrict a)
{
	switch (s) {
		case 0 :	/* hh */
			*va_arg(*a, signed char *) = v;
			break;
		
		case 1 :	/* h */
			*va_arg(*a, short *) = v;
			break;
		
		case 2 :
		default :
			*va_arg(*a, int *) = v;
			break;
		
		case 3 :	/* l */
			*va_arg(*a, long *) = v;
			break;
		
		case 4 :	/* ll, L */
			*va_arg(*a, long_long *) = v;
			break;
		
		case 5 :	/* z */
			*va_arg(*a, size_t *) = v;
			break;
		
		case 6 :	/* t */
			*va_arg(*a, ptrdiff_t *) = v;
			break;
		
		case 7 :	/* j */
			*va_arg(*a, intmax_t *) = v;
			break;
	}
}

static
int
handle_int(FILE * const restrict f, unsigned int const s, unsigned int * const restrict ss, int const assign,
va_list * const restrict a)
{
	intmax_t v;
	
	if (! *ss)
		*ss = BINIO_integer_sizes[s];
	
	if (get_int(f, &v, *ss))
		return 1;
	
	if (assign)
		BINIO_assign_int(v, s, a);
	
	return 0;
}

static
int
handle_uint(FILE * const restrict f, unsigned int const s, unsigned int * const restrict ss, int const assign,
va_list * const restrict a)
{
	uintmax_t v;
	
	if (! *ss)
		*ss = BINIO_integer_sizes[s];
	
	if (get_uint(f, &v, *ss))
		return 1;
	
	if (assign)
		switch (s) {
			case 0 :	/* hh */
				*va_arg(*a, unsigned char *) = v;
				break;
			
			case 1 :	/* h */
				*va_arg(*a, unsigned short *) = v;
				break;
			
			case 2 :
			default :
				*va_arg(*a, unsigned int *) = v;
				break;
			
			case 3 :	/* l */
				*va_arg(*a, unsigned long *) = v;
				break;
			
			case 4 :	/* ll, L */
				*va_arg(*a, unsigned_long_long *) = v;
				break;
			
			case 5 :	/* z */
				*va_arg(*a, size_t *) = v;
				break;
			
			case 6 :	/* t */
				*va_arg(*a, ptrdiff_t *) = v;
				break;
			
			case 7 :	/* j */
				*va_arg(*a, uintmax_t *) = v;
				break;
		}
	
	return 0;
}


static
int
handle_float(FILE * const restrict f, unsigned int const s, unsigned int * const restrict ss, int const assign,
va_list * const restrict a)
{
	long_double v;
	
	if (! ss[0])
		ss[0] = BINIO_float_sizes[s][0];
	if (! ss[1])
		ss[1] = BINIO_float_sizes[s][1];
	
	if (get_float(f, &v, ss[0], ss[1]))
		return 1;
	
	if (assign)
		switch (s) {
			case 0 :	/* hh */
			case 1 :	/* h */
			case 2 :
				*va_arg(*a, float *) = v;
				break;

			default :
			case 3 :	/* l */
				*va_arg(*a, double *) = v;
				break;
				
			case 4 :	/* ll, L */
			case 5 :	/* z */
			case 6 :	/* t */
			case 7 :	/* j */
				*va_arg(*a, long_double *) = v;
				break;
		}
	
	return 0;
}

#include <string.h>
#include <ctype.h>

static
int
handle_string(FILE * const restrict f, unsigned int const s, unsigned int const * const restrict ss, int const assign,
va_list * const restrict a, int const nul, char const * const restrict pat, int const include)
/*
 * FIXME: ranges in the pattern not implemented (note: this could quite possibly be never implemented.  using ranges
 * is implementation-defined behaviour, and, more importantly, just plain bad coding style)
 *
 */
{
	if (s > 2) {
		unsigned int i;
		wchar_t *p;
		
		for (i = 0, p = va_arg(*a, wchar_t *); ! *ss || (*ss && i < *ss); i++, p++) {
			wint_t c;
			
			c = BINIO_get_wchar(f);
			if (c == WEOF)
				return 1;
			else if (pat) {
				int const nc = wctob(c);
				
				if ((include && ! strchr(pat, nc))
				|| (! include && strchr(pat, nc)))
					break;
			} else if (iswspace(c))
				break;
			
			*p = c;
		}
		
		if (nul)
			*p = 0;
	} else {
		unsigned int i;
		char *p;
		
		for (i = 0, p = va_arg(*a, char *); ! *ss || (*ss && i < *ss); i++, p++) {
			int c;
			
			c = BINIO_get_char(f);
			if (c == EOF)
				return 1;
			else if (pat) {
				if ((include && ! strchr(pat, c))
				|| (! include && strchr(pat, c)))
					break;
			} else if (isspace(c))
				break;
			
			*p = c;
		}
		
		if (nul)
			*p = 0;
	}
	
	return 0;
}

static
int
handle_wstring(FILE * const restrict f, unsigned int const s, unsigned int const * const restrict ss, int const assign,
va_list * const restrict a, int const nul, wchar_t const * const restrict pat, int const include)
/*
 * FIXME: horrible redundancies between this and handle_string()
 *
 */
{
	if (s > 2) {
		unsigned int i;
		wchar_t *p;
		
		for (i = 0, p = va_arg(*a, wchar_t *); ! *ss || (*ss && i < *ss); i++, p++) {
			wint_t c;
			
			c = BINIO_get_wchar(f);
			if (c == WEOF)
				return 1;
			else if (pat) {
				if ((include && ! wcschr(pat, c))
				|| (! include && wcschr(pat, c)))
					break;
			} else if (iswspace(c))
				break;
			
			*p = c;
		}
		
		if (nul)
			*p = 0;
	} else {
		unsigned int i;
		char *p;
		
		for (i = 0, p = va_arg(*a, char *); ! *ss || (*ss && i < *ss); i++, p++) {
			int c;
			
			c = BINIO_get_char(f);
			if (c == EOF)
				return 1;
			else if (pat) {
				wint_t const nc = btowc(c);
			
				if ((include && ! wcschr(pat, nc))
				|| (! include && wcschr(pat, nc)))
					break;
			} else if (isspace(c))
				break;
			
			*p = c;
		}
		
		if (nul)
			*p = 0;
	}
	
	return 0;
}

static
int
handle_pattern(FILE * const restrict f, unsigned int const s, unsigned int const * const restrict ss, int const assign,
va_list * const restrict a, char const * restrict * const fmt, int const include)
{
	char const * const pat_start = *fmt;
	
	for (++*fmt; **fmt != ']'; ++*fmt);
	
	*fmt = 0;
	
	{
	int const r = handle_string(f, s, ss, assign, a, 1, pat_start, include);
	++*fmt;
	return r;
	}
}

static
int
handle_wpattern(FILE * const restrict f, unsigned int const s, unsigned int const * const restrict ss, int const assign,
va_list * const restrict a, wchar_t const * restrict * const fmt, int const include)
{
	wchar_t const * const pat_start = *fmt;
	
	for (++*fmt; **fmt != L']'; ++*fmt);
	
	*fmt = 0;

	{
	int const r = handle_wstring(f, s, ss, assign, a, 1, pat_start, include);
	++*fmt;
	return r;
	}
}

#include <ctype.h>
#include <stdlib.h>

int
vfgetf(FILE * const restrict f, char const * restrict fmt, va_list a)
{
	int count = 0;
	
	while (*fmt) {
		unsigned int size = 2;	/* 2 is normal size */
		unsigned int specific_sizes[2] = { 0, 0 };
		unsigned int specific_size_index = 0;
		int assign = 1;
		
		while (*fmt && *fmt++ != '%') {
			int c;
			
			c = getc(f);
			if (c == EOF || c != *fmt)
				goto failed;
		}
		
		if (*fmt == '%') {
			int c;
			
			c = getc(f);
			if (c == EOF || c != *fmt)
				goto failed;
			
			fmt++;
			continue;
		}
		
		while (*fmt) {
			if (*fmt != '0' && isdigit(*fmt)) {
				specific_sizes[specific_size_index] = strtoul(fmt, (char **)&fmt, 10);
				continue;
			}
			
			switch (*fmt) {
				default :
					/* ignore */
					break;
				
				case '.' :
					specific_size_index++;
					break;
				
				case '*' :
					assign = 0;
					break;

				case 'h' :
					size--;
					break;
				
				case 'l' :
					size++;
					break;
				
				case 'L' :
					size = 4;
					break;

				case 'z' :
					size = 5;
					break;

				case 't' :
					size = 6;
					break;
				
				case 'j' :
					size = 7;
					break;
				
				case 'd' :
				case 'i' :
					if (handle_int(f, size, specific_sizes, assign, &a))
						goto failed;
					if (assign)
						count++;
					fmt++;
					goto end_loop;
				
				case 'o' :
				case 'u' :
				case 'x' :
					if (handle_uint(f, size, specific_sizes, assign, &a))
						goto failed;
					if (assign)
						count++;
					fmt++;
					goto end_loop;
				
				case 'a' :
				case 'e' :
				case 'f' :
				case 'g' :
					if (handle_float(f, size, specific_sizes, assign, &a))
						goto failed;
					if (assign)
						count++;
					fmt++;
					goto end_loop;
				
				case 'c' :
					if (! specific_sizes[0])
						specific_sizes[0] = 1;
					if (handle_string(f, size, specific_sizes, assign, &a, 0, 0, 0))
						goto failed;
					if (assign)
						count++;
					fmt++;
					goto end_loop;
				
				case 's' :
					if (handle_string(f, size, specific_sizes, assign, &a, 1, 0, 0))
						goto failed;
					if (assign)
						count++;
					fmt++;
					goto end_loop;
				
				case '[' :
					if ((*++fmt == '^' && (fmt++,
					handle_pattern(f, size, specific_sizes, assign, &a, &fmt, 0)))
					|| handle_pattern(f, size, specific_sizes, assign, &a, &fmt, 1))
						goto failed;
					if (assign)
						count++;
					goto end_loop;
				
				case 'n' :
					BINIO_assign_int(count, size, &a);
					fmt++;
					goto end_loop;
			}
			fmt++;
		}
		
		end_loop : ;
	}
	
	return count;
	
	failed :
	
	return count ? count : EOF;
}

#ifdef BINIO_HAVE_WCHAR

int
vfwgetf(FILE * const restrict f, wchar_t const * restrict fmt, va_list a)
/*
 * FIXME: great redundancy between this function and vfgetf
 *
 */
{
	int count = 0;
	
	while (*fmt) {
		unsigned int size = 2;	/* 2 is normal size */
		unsigned int specific_sizes[2] = { 0, 0 };
		unsigned int specific_size_index = 0;
		int assign = 1;
		
		while (*fmt && *fmt++ != L'%') {
			wint_t c;
			
			c = BINIO_get_wchar(f);
			if (c == WEOF || c != *fmt)
				goto failed;
		}
		
		if (*fmt == L'%') {
			wint_t c;
			
			c = BINIO_get_wchar(f);
			if (c == WEOF || c != *fmt)
				goto failed;
			
			fmt++;
			continue;
		}
		
		while (*fmt) {
			if (*fmt != L'0' && iswdigit(*fmt)) {
				specific_sizes[specific_size_index] = wcstoul(fmt, (wchar_t **)&fmt, 10);
				continue;
			}
			
			switch (*fmt) {
				default :
					/* ignore */
					break;
				
				case L'.' :
					specific_size_index++;
					break;
				
				case L'*' :
					assign = 0;
					break;

				case L'h' :
					size--;
					break;
				
				case 'l' :
					size++;
					break;
				
				case 'L' :
					size = 4;
					break;

				case L'z' :
					size = 5;
					break;

				case L't' :
					size = 6;
					break;
				
				case L'j' :
					size = 7;
					break;
				
				case L'd' :
				case L'i' :
					if (handle_int(f, size, specific_sizes, assign, &a))
						goto failed;
					if (assign)
						count++;
					fmt++;
					goto end_loop;
				
				case L'o' :
				case L'u' :
				case L'x' :
					if (handle_uint(f, size, specific_sizes, assign, &a))
						goto failed;
					if (assign)
						count++;
					fmt++;
					goto end_loop;
				
				case L'a' :
				case L'e' :
				case L'f' :
				case L'g' :
					if (handle_float(f, size, specific_sizes, assign, &a))
						goto failed;
					if (assign)
						count++;
					fmt++;
					goto end_loop;
				
				case L'c' :
					if (! specific_sizes[0])
						specific_sizes[0] = 1;
					if (handle_string(f, size, specific_sizes, assign, &a, 0, 0, 0))
						goto failed;
					if (assign)
						count++;
					fmt++;
					goto end_loop;
				
				case L's' :
					if (handle_string(f, size, specific_sizes, assign, &a, 1, 0, 0))
						goto failed;
					if (assign)
						count++;
					fmt++;
					goto end_loop;
				
				case L'[' :
					if ((*++fmt == L'^' && (fmt++,
					handle_wpattern(f, size, specific_sizes, assign, &a, &fmt, 0)))
					|| handle_wpattern(f, size, specific_sizes, assign, &a, &fmt, 1))
						goto failed;
					if (assign)
						count++;
					goto end_loop;
				
				case L'n' :
					BINIO_assign_int(count, size, &a);
					fmt++;
					goto end_loop;
			}
			fmt++;
		}
		
		end_loop : ;
	}
	
	return count;
	
	failed :
	
	return count ? count : EOF;
}

#endif
