#include "binio-internal.h"
#include "binio.h"
#include <stdio.h>
#include <stdint.h>

static
int
put_uint(FILE * const restrict f, uintmax_t const n, unsigned int const s)
{
	unsigned int i;
	unsigned int const rs = s < sizeof n ? s : sizeof n;
	
	for (i = rs; i < s; i++)
		if (putc(0, f) == EOF)
			return i;
		/*
		 * put leading zeroes if uintmax_t is too small to handle what we're dealing with.  see get.c:get_uint
		 * for a better comment describing this
		 *
		 */
	for (i = 0; i < rs; i++)
		if (putc((n >> (BYTE_BIT * (s - i - 1))) & BYTE_MASK, f) == EOF)
			break;
	
	return i + s - rs;
	
}

static
int
put_int(FILE * const restrict f, intmax_t const n, unsigned int const s)
{
	return put_uint(f, n < 0 ? ~(uintmax_t)-n + 1 : n, s);
}

#include <math.h>
#include <limits.h>

static
int
put_float(FILE * const restrict f, long_double const v, unsigned int const mantissa_size, unsigned int const exponent_size)
{
	int exponent;
	size_t const ms_bit_size = BYTE_BIT * mantissa_size;
	size_t shift_amount = ms_bit_size > sizeof (intmax_t) * CHAR_BIT ? sizeof (intmax_t) * CHAR_BIT : ms_bit_size;
	long_double const fraction = frexpl(v, &exponent);
	intmax_t const mantissa = ldexpl(fraction, shift_amount - 1);
	int r, cr;
	
	r = put_int(f, mantissa, mantissa_size);
	if (r < mantissa_size)	return r;
	cr = put_int(f, exponent, exponent_size);
	return r + cr;
}

#include <stdarg.h>

int
fputf(FILE * const restrict f, char const * const restrict fmt, ...)
{
	va_list a;
	
	va_start(a, fmt);
	{
	int const r = vfputf(f, fmt, a);
	va_end(a);
	
	return r;
	}
}

#ifdef BINIO_HAVE_WCHAR

int
fwputf(FILE * const restrict f, wchar_t const * const restrict fmt, ...)
{
	va_list a;
	
	va_start(a, fmt);
	{
	int const r = vfwgetf(f, fmt, a);
	va_end(a);
	
	return r;
	}
}

#endif

static
int
handle_float(FILE * const restrict f, unsigned int const s, unsigned int * const restrict ss, va_list * const restrict a)
{
	long_double n;

	switch (s) {
		case 0 :	/* hh */
		case 1 :	/* h */
		case 2 :
		default :
		case 3 :	/* l */
			n = va_arg(*a, double);
			break;
			
		case 4 :	/* ll, L */
		case 5 :	/* z */
		case 6 :	/* t */
		case 7 :	/* j */
			n = va_arg(*a, long_double);
			break;
	}

	if (! ss[0])
		ss[0] = BINIO_float_sizes[s][0];
	if (! ss[1])
		ss[1] = BINIO_float_sizes[s][1];
	
	return put_float(f, n, ss[0], ss[1]);
}

#include <stddef.h>

static
int
handle_uinteger(FILE * const restrict f, unsigned int s, unsigned int * const restrict ss, va_list * const restrict a)
{
	uintmax_t n;
	
	switch (s) {
		case 0 :	/* hh */
			if (sizeof (char) < sizeof (int)) {
				n = va_arg(*a, int);
				break;
			}
			/*
			 * if sizeof (char) == sizeof (int), then the argument is coming as an unsigned int, so we
			 * deliberately fall through here
			 *
			 */
		
		case 1 :	/* h */
			if (sizeof (short) < sizeof (int)) {
				n = va_arg(*a, int);
				break;
			}
			/*
			 * similarly to above, we deliberately fall through
			 *
			 */
		
		case 2 :
		default :
			n = va_arg(*a, unsigned int);
			break;

		case 3 :	/* l */
			n = va_arg(*a, unsigned long);
			break;
			
		case 4 :	/* ll, L */
			n = va_arg(*a, unsigned_long_long);
			break;
		
		case 5 :	/* z */
			n = (intmax_t)va_arg(*a, size_t);
			break;
		
		case 6 :	/* t */
			n = va_arg(*a, ptrdiff_t);
			break;

		case 7 :	/* j */
			n = va_arg(*a, uintmax_t);
			break;
	}

	if (*ss)
		s = *ss;
	else
		s = BINIO_integer_sizes[s];
	
	return put_uint(f, n, s);
}

static
int
handle_integer(FILE * const restrict f, unsigned int s, unsigned int * const restrict ss, va_list * const restrict a)
{
	intmax_t n;
	
	switch (s) {
		case 0 :	/* hh */
		case 1 :	/* h */
		case 2 :
		default :
			n = va_arg(*a, int);
			break;

		case 3 :	/* l */
			n = va_arg(*a, long);
			break;
			
		case 4 :	/* ll, L */
			n = va_arg(*a, long_long);
			break;
		
		case 5 :	/* z */
			n = va_arg(*a, size_t);
			break;
		
		case 6 :	/* t */
			n = va_arg(*a, ptrdiff_t);
			break;

		case 7 :	/* j */
			n = va_arg(*a, intmax_t);
			break;
	}
	
	if (*ss)
		s = *ss;
	else
		s = BINIO_integer_sizes[s];
	
	return put_int(f, n, s);
}

static
int
handle_char(FILE * const restrict f, unsigned int const s, va_list * const restrict a)
{
	if (s > 2)
		return BINIO_put_wchar(f, va_arg(*a, wchar_t));
	else
		return BINIO_put_char(f, va_arg(*a, int));
}

#include <limits.h>

static
int
handle_string(FILE * const restrict f, unsigned int const s, unsigned int * const restrict ss, va_list * const restrict a)
{
	if (! *ss)
		*ss = UINT_MAX;

	if (s > 2) {
		wchar_t const * restrict p;
		int c = 0;
		unsigned int i;
		
		for (i = 0, p = va_arg(*a, wchar_t const *); *p && i < *ss; p++, i++) {
			int cr;
			
			cr = BINIO_put_wchar(f, *p);
			c += cr;
			if (! cr)
				return c;
		}
		
		return c;
	} else {
		char const * restrict p;
		int c = 0;
		unsigned int i;
		
		for (i = 0, p = va_arg(*a, char const *); *p && i < *ss; p++, i++) {
			int cr;
			
			cr = BINIO_put_char(f, *p);
			c += cr;
			if (! cr)
				return c;
		}
		
		return c;
	}
}

#include <ctype.h>
#include <stdlib.h>

int
vfputf(FILE * const restrict f, char const * restrict fmt, va_list a)
{
	int count = 0;
	
	while (*fmt) {
		unsigned int size = 2;	/* 2 is normal size */
		unsigned int specific_sizes[2] = { 0, 0 };
		unsigned int specific_size_index = 0;
		
		while (*fmt && *fmt++ != '%') {
			if (fputc(*fmt, f) == EOF)
				break;
			count++;
		}
		
		if (*fmt == '%') {
			if (fputc('%', f) == EOF)
				break;
			count++;
			fmt++;
			continue;
		}
		
		while (*fmt) {
			if (*fmt != '0' && isdigit(*fmt)) {
				specific_sizes[specific_size_index] = strtoul(fmt, (char **)&fmt, 10);
				continue;
			}
			
			switch (*fmt) {
				int cr;
				
				case '0' :
				case '-' :
				case '+' :
				case ' ' :
				case '#' :
				default :
					/* ignore */
					break;
				
				case '.' :
					specific_size_index++;
					break;
				
				case '*' :
					specific_sizes[specific_size_index] = va_arg(a, int);
					break;
				
				case 'h' :
					size--;
					break;
				
				case 'l' :
					size++;
					break;
				
				case 'L' :
					size = 4;
					break;

				case 'z' :
					size = 5;
					break;

				case 't' :
					size = 6;
					break;
				
				case 'j' :
					size = 7;
					break;
				
				case 'd' :
				case 'i' :
					cr = handle_integer(f, size, specific_sizes, &a);
					if (! cr)
						goto failed;
					count += cr;
					fmt++;
					goto end_loop;
				
				case 'u' :
				case 'o' :
				case 'x' :
				case 'X' :
					cr = handle_uinteger(f, size, specific_sizes, &a);
					if (! cr)
						goto failed;
					count += cr;
					fmt++;
					goto end_loop;
				
				case 'f' :
				case 'F' :
				case 'e' :
				case 'E' :
				case 'g' :
				case 'G' :
				case 'a' :
				case 'A' :
					cr = handle_float(f, size, specific_sizes, &a);
					if (! cr)
						goto failed;
					count += cr;
					fmt++;
					goto end_loop;
				
				case 'c' :
					cr = handle_char(f, size, &a);
					if (! cr)
						goto failed;
					count += cr;
					fmt++;
					goto end_loop;
				
				case 's' :
					cr = handle_string(f, size, specific_sizes, &a);
					if (! cr)
						goto failed;
					count += cr;
					fmt++;
					goto end_loop;
				
				case 'n' :
					BINIO_assign_int(count, size, &a);
					fmt++;
					goto end_loop;
			}
			fmt++;
		}
		
		end_loop : ;
	}
	
	failed :
	
	return count;
}

#ifdef BINIO_HAVE_WCHAR

int
vfwputf(FILE * const restrict f, wchar_t const * restrict fmt, va_list a)
/*
 * FIXME: redundancy between this and vfputf()
 *
 */
{
	int count = 0;
	
	while (*fmt) {
		unsigned int size = 2;	/* 2 is normal size */
		unsigned int specific_sizes[2] = { 0, 0 };
		unsigned int specific_size_index = 0;
		
		while (*fmt && *fmt++ != L'%') {
			if (BINIO_put_wchar(f, *fmt) == WEOF)
				break;
			count++;
		}
		
		if (*fmt == L'%') {
			if (BINIO_put_wchar(f, L'%') == WEOF)
				break;
			count++;
			fmt++;
			continue;
		}
		
		while (*fmt) {
			if (*fmt != L'0' && iswdigit(*fmt)) {
				specific_sizes[specific_size_index] = wcstoul(fmt, (wchar_t **)&fmt, 10);
				continue;
			}
			
			switch (*fmt) {
				int cr;
				
				case L'0' :
				case L'-' :
				case L'+' :
				case L' ' :
				case L'#' :
				default :
					/* ignore */
					break;
				
				case L'.' :
					specific_size_index++;
					break;
				
				case L'*' :
					specific_sizes[specific_size_index] = va_arg(a, int);
					break;
				
				case L'h' :
					size--;
					break;
				
				case L'l' :
					size++;
					break;
				
				case L'L' :
					size = 4;
					break;

				case L'z' :
					size = 5;
					break;

				case L't' :
					size = 6;
					break;
				
				case L'j' :
					size = 7;
					break;
				
				case L'd' :
				case L'i' :
					cr = handle_integer(f, size, specific_sizes, &a);
					if (! cr)
						goto failed;
					count += cr;
					fmt++;
					goto end_loop;
				
				case L'u' :
				case L'o' :
				case L'x' :
				case L'X' :
					cr = handle_uinteger(f, size, specific_sizes, &a);
					if (! cr)
						goto failed;
					count += cr;
					fmt++;
					goto end_loop;
				
				case L'f' :
				case L'F' :
				case L'e' :
				case L'E' :
				case L'g' :
				case L'G' :
				case L'a' :
				case L'A' :
					cr = handle_float(f, size, specific_sizes, &a);
					if (! cr)
						goto failed;
					count += cr;
					fmt++;
					goto end_loop;
				
				case L'c' :
					cr = handle_char(f, size, &a);
					if (! cr)
						goto failed;
					count += cr;
					fmt++;
					goto end_loop;
				
				case L's' :
					cr = handle_string(f, size, specific_sizes, &a);
					if (! cr)
						goto failed;
					count += cr;
					fmt++;
					goto end_loop;
				
				case L'n' :
					BINIO_assign_int(count, size, &a);
					fmt++;
					goto end_loop;
			}
			fmt++;
		}
		
		end_loop : ;
	}
	
	failed :
	
	return count;
}


#endif
